{
  Additional memory manager, speeding up memory allocation and reallocation
  operations. It uses more memory, but allocates it for the
  application faster, and especially optimizes reallocation operations.
  =====================================================================
  Copyright (C) by Vladimir Kladov, 2005
  ---------------------------------------------------------------------
  http://bonanzas.rinet.ru
  mailto: bonanzas@online.sinor.ru
}

unit FastMM;

{$OPTIMIZATION ON}
{$STACKFRAMES OFF}

interface

{$I KOLDEF.INC}

uses windows, KOL;

{$IFNDEF NOTUSE_CRITICALSECTION}
{$DEFINE USE_CRITICALSECTION} // for single-thread applications, it is possible
                              // to comment this declaration (or just add
                              // conditional symbol NOTUSE_CRITICALSECTION in
                              // project options to work even faster).
{$ENDIF NOTUSE_CRITICALSECTION}

{$IFNDEF NOASM_CODE}
{$DEFINE ASM_CODE}
{$ENDIF  NOASM_CODE}

const MaxBlockSz = $100000; // Maximum block allocated with FastMM, any larger
                            // block is allocated using old memory manager.

procedure Install_FastMM;
// this procedure can be called at any time, not only in initialization block
// and after it is called, new memory manager starts functioning.

implementation

//------------------------------------------------------------------------------
// Memory block structure
//------------------------------------------------------------------------------
type
  PBlockInfo = ^TBlockInfo;
  TBlockInfo = packed record
    prev: PBlockInfo;
    //next: PBlockInfo;
    smaller: PBlockInfo;
    larger: PBlockInfo;
    freesz: Integer;
    allocsz: Integer;
  end;

//------------------------------------------------------------------------------
// Index of blocks sorted by free size: contains pointers to blocks of largest
// size in the same granularity group of the size.
//------------------------------------------------------------------------------
var
  BlockStarts, BlockEnds: array[ 0..20 ] of PBlockInfo;
  // and some statistics here:
  {$IFDEF STATS}
  Allocated_Blocks: Integer;
  Allocated_FromEmpty: Integer;
  Allocated_Dividing: Integer;
  Reallocated: Integer;
  Freed: Integer;
  {$ENDIF STATS}

//------------------------------------------------------------------------------
// Usage statistics
//------------------------------------------------------------------------------
var BlocksAllocated, BlocksFree: array[ 0..20 ] of Integer;

//------------------------------------------------------------------------------
// Install FastMM memory manager
//------------------------------------------------------------------------------
function GetMem_FastMM(size: Integer): Pointer; forward;
function FreeMem_FastMM(p: Pointer): Integer; forward;
function ReallocMem_FastMM(p: Pointer; size: Integer): Pointer; forward;

{$IFDEF USE_CRITICALSECTION}
var CritSec_FastMM: TRTLCriticalSection;
{$ENDIF}
var OldMM: TMemoryManager;
const
  FastMemoryManager: TMemoryManager = (
    GetMem    : GetMem_FastMM;
    FreeMem   : FreeMem_FastMM;
    ReallocMem: ReallocMem_FastMM
  );

var FastMM_Installed: Boolean = FALSE;
    Uninitialization: Boolean = FALSE;
procedure Install_FastMM;
begin
  if FastMM_Installed then Exit;
  {$IFDEF USE_CRITICALSECTION}
  InitializeCriticalSection( CritSec_FastMM );
  EnterCriticalSection( CritSec_FastMM );
  TRY
  {$ENDIF USE_CRITICALSECTION}
    GetMemoryManager( OldMM );
    SetMemoryManager( FastMemoryManager );
    FastMM_Installed := TRUE;
  {$IFDEF USE_CRITICALSECTION}
  FINALLY
    LeaveCriticalSection( CritSec_FastMM );
  END;
  {$ENDIF USE_CRITICALSECTION}
end;

function log( s: PChar; B: PBlockInfo ): Boolean;
begin
  Result := TRUE;
  SetMemoryManager( OldMM );
  TRY
    LogFileOutput( GetStartDir + 'FastMM.log',
      s + #9 + Int2Hex( Integer( B ), 6 ) +
      ' free:' + Int2Str( B.freesz ) +
      ' alloc: ' + Int2Str( B.allocsz xor -1 ) +
      ' smaller: ' + Int2Hex( Integer( B.smaller ), 6 ) +
      ' larger: ' + Int2Hex( Integer( B.larger ), 6 ) );
  FINALLY
    SetMemoryManager( FastMemoryManager );
  END;
end;

//------------------------------------------------------------------------------
// Service functions to store all blocks sorted by freesz in descending order
//------------------------------------------------------------------------------
function BlockIndex( freesz: Integer ): Integer;
begin
  {$IFDEF PASCAL_ONLY}
  if freesz <= Sizeof( TBlockInfo ) + 4 then
    Result := -1
  else
  CASE freesz OF
  1: Result := 0;
  2: Result := 1;
  3,4: Result := 2;
  5..8: Result := 3;
  9..16: Result := 4;
  17..32: Result := 5;
  33..64: Result := 6;
  65..128: Result := 7;
  128..256: Result := 8;
  257..512: Result := 9;
  513..1024: Result := 10;
  1025..2048: Result := 11;
  2049..4096: Result := 12;
  4097..8192: Result := 13;
  8193..16384: Result := 14;
  16383..32768: Result := 15;
  32768..65536: Result := 16;
  65537..$2000: Result := 17;
  $2001..$4000: Result := 18;
  $4001..$8000: Result := 19;
  else Result := 20;
  END;
  {$ELSE}
  if freesz <= Sizeof( TBlockInfo ) + 4 then
    Result := -1
  else
  begin
    asm
      bsr eax, [freesz]
      mov @Result, eax
    end;
    if 1 shl Result < freesz then
      Inc( Result );
    if Result > High( BlockStarts ) then
      Result := High( BlockStarts );
  end;
  {$ENDIF}
end;

{$IFDEF ASM_CODE}
procedure Unlink( B: PBlockInfo );
const C1 = Sizeof( TBlockInfo ) + 4;
      Bsz = Sizeof( TBlockInfo );
asm
  push ebx
  xchg ebx, eax
  xor ecx, ecx
  dec ecx
  cmp dword ptr [ebx].TBlockInfo.freesz, C1
  jle @@2
  bsr ecx, [ebx].TBlockInfo.freesz
  xor edx, edx
  inc edx
  shl edx, cl
  cmp edx, [ebx].TBlockInfo.freesz
  jge @@1
  inc ecx
@@1:
  cmp ecx, 20
  jle @@2
  mov ecx, 20
@@2:
  test ecx, ecx
  jle @@3
  mov edx, [ebx].TBlockInfo.allocsz
  not edx
  add edx, Bsz
  sub dword ptr [ecx*4+BlocksAllocated], edx
  mov edx, [ebx].TBlockInfo.freesz
  sub dword ptr [ecx*4+BlocksFree], edx
@@3:
  mov edx, [ebx].TBlockInfo.smaller
  mov eax, [ebx].TBlockInfo.larger
  test edx, edx
  jz  @@4
  mov  [edx].TBlockInfo.larger, eax
@@4:
  test eax, eax
  jz  @@5
  mov  [eax].TBlockInfo.smaller, edx
@@5:
  test ecx, ecx
  jl  @@8
  cmp dword ptr [ecx*4+BlockStarts], ebx
  jne @@6
  mov dword ptr [ecx*4+BlockStarts], edx
@@6:
  cmp dword ptr [ecx*4+BlockEnds], ebx
  jne @@7
  mov dword ptr [ecx*4+BlockEnds], eax
@@7:
  xor eax, eax
  cmp dword ptr [ecx*4+BlockStarts], eax
  jne @@8
  cmp [Uninitialization], al
  jz  @@8
@@loo9:
  cmp [eax*4+BlockStarts], 0
  jnz @@8
  inc eax
  cmp eax, 20
  jle @@loo9
  {$IFDEF USE_CRITICALSECTION}
  lea eax, [CritSec_FastMM]
  push eax
  call DeleteCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  lea  eax, [OldMM]
  call SetMemoryManager
  mov  [FastMM_Installed], 0
@@8:
  xor eax, eax
  mov [ebx].TBlockInfo.smaller, eax
  mov [ebx].TBlockInfo.larger, eax
  pop ebx
end;
{$ELSE}
procedure Unlink( B: PBlockInfo );
var i: Integer;
    All0: Boolean;
begin
  //Assert( log( 'unlink', B ) );
  i := BlockIndex( B.freesz );
  if i >= 0 then
  begin
    Dec( BlocksAllocated[ i ], (B.allocsz xor -1) + Sizeof( B^ ) );
    Dec( BlocksFree[ i ], B.freesz );
  end;
  if B.smaller <> nil then
    B.smaller.larger := B.larger;
  if B.larger <> nil then
  begin
    B.larger.smaller := B.smaller;
    {if B.larger = B.smaller then
    asm
      int 3
    end;}
  end;
  if i >= 0 then
  begin
    if BlockStarts[ i ] = B then
      BlockStarts[ i ] := B.smaller;
    if BlockEnds[ i ] = B then
      BlockEnds[ i ] := B.larger;
    if (BlockStarts[ i ] = nil) and Uninitialization then
    begin
      All0 := TRUE;
      for i := 0 to High( BlockStarts ) do
        if (BlockStarts[ i ] <> nil) or (BlockEnds[ i ] <> nil) then
        begin
          All0 := FALSE;
          break;
        end;
      if All0 then
      begin
        {$IFDEF USE_CRITICALSECTION}
        DeleteCriticalSection( CritSec_FastMM );
        {$ENDIF USE_CRITICALSECTION}
        SetMemoryManager( OldMM );
        FastMM_Installed := FALSE;
      end;
    end;
  end;
  B.smaller := nil;
  B.larger := nil;
end;
{$ENDIF NOASM_CODE}

{$IFDEF ASM_CODE}
procedure InsertSorted( B: PBlockInfo );
const C1 = Sizeof( TBlockInfo ) + 4;
      Bsz = Sizeof( TBlockInfo );
asm
  push ebx
  xchg eax, ebx
  xor ecx, ecx
  dec ecx
  cmp dword ptr [ebx].TBlockInfo.freesz, C1
  jle @@2
  bsr ecx, [ebx].TBlockInfo.freesz
  xor edx, edx
  inc edx
  shl edx, cl
  cmp edx, [ebx].TBlockInfo.freesz
  jge @@1
  inc ecx
@@1:
  cmp ecx, 20
  jle @@2
  mov ecx, 20
@@2:
  test ecx, ecx
  jge  @@3
  xor  eax, eax
  mov  [ebx].TBlockInfo.smaller, eax
  mov  [ebx].TBlockInfo.larger, eax
  jmp  @@exit
@@3:
  mov  edx, [ebx].TBlockInfo.allocsz
  not  edx
  add  edx, Bsz
  add  dword ptr [ecx*4+BlocksAllocated], edx
  mov  edx, [ebx].TBlockInfo.freesz
  add  dword ptr [ecx*4+BlocksFree], edx
  mov  eax, dword ptr [ecx*4+BlockStarts]
  test eax, eax
  jnz @@4
  mov  dword ptr [ecx*4+BlockStarts], ebx
  mov  dword ptr [ecx*4+BlockEnds], ebx
  mov  [ebx].TBlockInfo.smaller, eax
  mov  [ebx].TBlockInfo.larger, eax
  jmp  @@exit
@@4:
  mov  eax, [ebx].TBlockInfo.allocsz
  not  eax
  mov  edx, [ebx].TBlockInfo.freesz
  shl  edx, 3
  cmp  eax, edx
  jge  @@5
  mov  eax, dword ptr [ecx*4+BlockStarts]
  mov  [ebx].TBlockInfo.smaller, eax
  mov  [eax].TBlockInfo.larger, ebx
  mov  dword ptr [ecx*4+BlockStarts], ebx
  cmp  dword ptr [ecx*4+BlockEnds], 0
  jnz  @@exit
  mov  dword ptr [ecx*4+BlockEnds], ebx
  jmp  @@exit
@@5:
  mov  eax, dword ptr [ecx*4+BlockEnds]
  mov  [ebx].TBlockInfo.larger, eax
  test eax, eax
  jz   @@6
  mov  [eax].TBlockInfo.smaller, ebx
@@6:
  //cmp  dword ptr [ecx*4+BlockStarts], 0
  //jnz  @@exit
  //mov  dword ptr [ecx*4+BlockStarts], ebx
@@exit:
  pop  ebx
end;
{$ELSE}
procedure InsertSorted( B: PBlockInfo );
var i: Integer;
    B1: PBlockInfo;
begin
  //Assert( log( 'insert', B ) );
  i := BlockIndex( B.freesz );
  if i < 0 then
  begin
    B.smaller := nil;
    B.larger := nil;
    Exit;
  end;
  Inc( BlocksAllocated[ i ], (B.allocsz xor -1) + Sizeof( B^ ) );
  Inc( BlocksFree[ i ], B.freesz );
  if BlockStarts[ i ] = nil then
  begin
    BlockStarts[ i ] := B;
    BlockEnds[ i ] := B;
    B.smaller := nil;
    B.larger := nil;
  end
    else
  begin
    if (B.allocsz xor -1) < B.freesz * 8 then
    begin
      B1 := BlockStarts[ i ];

      B.smaller := B1;
      //if B1 <> nil then
        B1.larger := B;
      BlockStarts[ i ] := B;

      if BlockEnds[ i ] = nil then
        BlockEnds[ i ] := B;
    end
      else
    begin
      B1 := BlockEnds[ i ];

      B.larger := B1;
      if B1 <> nil then
        B1.smaller := B;
      BlockEnds[ i ] := B;

      {if BlockStarts[ i ] = nil then
        BlockStarts[ i ] := B;}
    end;

    {if BlockEnds[ i ] = nil then
      BlockEnds[ i ] := B;
    B1 := BlockStarts[ i ];
    while B1 <> nil do
    begin
      if B.freesz >= B1.freesz div 2 then
      begin
        B.smaller := B1;
        B.larger := B1.larger;
        B1.larger := B;
        if B.larger = nil then
          BlockStarts[ i ] := B
        else
          B.larger.smaller := B;
        Exit;
      end;
      if B1.smaller = nil then
      begin
        B1.smaller := B;
        B.larger := B1;
        B.smaller := nil;
        Exit;
      end;
      B1 := B1.smaller;
    end;}
  end;
end;
{$ENDIF NOASM_CODE}

//------------------------------------------------------------------------------
// Allocate memory block of given size
//------------------------------------------------------------------------------
{$IFDEF ASM_CODE}
function GetMem_FastMM(size: Integer): Pointer;
const Bsz = Sizeof( TBlockInfo );
asm
  push ebx
  push esi
  push edi
  xchg esi, eax
  xor  eax, eax
  cmp  esi, eax
  jle  @@exit
  cmp  esi, MaxBlockSz
  jl   @@1
  xchg eax, esi
  call dword ptr [OldMM.GetMem]
  jmp  @@exit
@@1:
  {$IFDEF USE_CRITICALSECTION}
  push offset [CritSec_FastMM]
  call EnterCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  xor ebx, ebx
  lea edi, [esi*2+(3+Bsz)*2]
  and edi, not 7
  bsr ecx, edi
  xor edx, edx
  inc edx
  shl edx, cl
  cmp edx, edi
  jge @@2
  inc ecx
@@2:
  cmp ecx, 20
  jle @@3
  mov ecx, 20
@@3:
  cmp [ecx*4+BlockStarts], 0
  jz  @@4
  mov edx, [ecx*4+BlocksFree]
  add edx, edx
  cmp [ecx*4+BlocksAllocated], edx
  jg  @@4
  mov ebx, [ecx*4+BlockStarts]
  mov eax, ebx
  call Unlink
  jmp @@5
@@4:
  inc ecx
  cmp ecx, 20
  jle @@3
@@5:
  test ebx, ebx
  jnz  @@6
  {$IFDEF STATS}
  inc  [Allocated_Blocks]
  {$ENDIF STATS}
  mov  eax, edi
  shl eax, 6
  or  eax, $FFFFF
  inc eax
  and eax, not $FFFFF
  mov edi, eax
  call dword ptr [OldMM.GetMem]
  xchg ebx, eax
  xor  eax, eax
  mov  [ebx].TBlockInfo.prev, eax
  mov  [ebx].TBlockInfo.smaller, eax
  mov  [ebx].TBlockInfo.larger, eax
  lea  edx, [edi-2*Bsz]
  mov  [ebx].TBlockInfo.freesz, edx
  or   [ebx].TBlockInfo.allocsz, -1
  lea  edx, [ebx+edx+Bsz]
  mov  [edx].TBlockInfo.smaller, eax
  mov  [edx].TBlockInfo.larger, eax
  mov  [edx].TBlockInfo.freesz, eax
  mov  [edx].TBlockInfo.prev, ebx
  or   [edx].TBlockInfo.allocsz, -1
@@6:
  cmp  [ebx].TBlockInfo.allocsz, -1
  jnz  @@7
  {$IFDEF STATS}
  inc  [Allocated_FromEmpty]
  {$ENDIF STATS}
  mov  eax, esi
  not  eax
  mov  [ebx].TBlockInfo.allocsz, eax
  sub  [ebx].TBlockInfo.freesz, esi
  jmp  @@9
@@7:
  {$IFDEF STATS}
  inc  [Allocated_Dividing]
  {$ENDIF STATS}
  mov  eax, [ebx].TBlockInfo.freesz
  shr  eax, 1
  lea  ecx, [esi+Bsz]
  cmp  eax, ecx
  jge  @@8
  xchg eax, ecx
@@8:
  add  eax, 3
  and  eax, not 3
  mov  edx, [ebx].TBlockInfo.allocsz
  not  edx
  lea  edi, [ebx+edx+Bsz]
  add  edi, [ebx].TBlockInfo.freesz
  sub  edi, eax
  and  [edi].TBlockInfo.smaller, 0
  and  [edi].TBlockInfo.larger, 0
  mov  [edi].TBlockInfo.prev, ebx
  sub  [ebx].TBlockInfo.freesz, eax
  lea  edx, [eax-Bsz]
  sub  edx, esi
  mov  [edi].TBlockInfo.freesz, edx
  mov  ecx, esi
  not  ecx
  mov  [edi].TBlockInfo.allocsz, ecx
  lea  edx, [edi+edx+Bsz]
  add  edx, esi
  mov  [edx].TBlockInfo.prev, edi
  xchg eax, ebx
  mov  ebx, edi
  call InsertSorted
@@9:
  mov  eax, ebx
  call InsertSorted
  {$IFDEF USE_CRITICALSECTION}
  push offset [CritSec_FastMM]
  call LeaveCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  lea  eax, [ebx+Bsz]
@@exit:
  pop  edi
  pop  esi
  pop  ebx
end;
{$ELSE}
function GetMem_FastMM(size: Integer): Pointer;
var B, B2, B3: PBlockInfo;
    blksize, cutsize: Integer;
    i, reqsize: Integer;
begin
  Assert( size >= 0, 'Attempt to allocate size < 0' );
  if size = 0 then
  begin
    Result := nil;
    Exit;
  end;
  if size >= MaxBlockSz then
  begin
    Result := OldMM.GetMem( size );
    Exit;
  end;
  {$IFDEF USE_CRITICALSECTION}
  EnterCriticalSection( CritSec_FastMM );
  TRY
  {$ENDIF USE_CRITICALSECTION}
    B := nil;
    reqsize := (size + 3 + Sizeof( B^ )) and not 3;
    i := BlockIndex( reqsize * 2 );
    for i := i to High( BlockStarts ) do
    begin
      if (BlockStarts[ i ] <> nil) and (BlocksAllocated[ i ] <= BlocksFree[ i ] div 2) then
      begin
        B := BlockStarts[ i ];
        Unlink( B );
        break;
      end;
    end;
    if B = nil then
    begin //         -   
      {$IFDEF STATS}
      Inc( Allocated_Blocks );
      {$ENDIF}
      blksize := ((reqsize * 128 or $FFFFF) + 1) and not $FFFFF;
      B := OldMM.GetMem( blksize );
      //FillChar( B^, Sizeof( B^ ), 0 );
      B.prev := nil;
      B.smaller := nil;
      B.larger := nil;
      B.freesz := blksize - Sizeof( B^ ) - Sizeof( B3^ );
      B.allocsz := -1;
      B3 := Pointer( Integer( B ) + Sizeof( B^ ) + B.freesz + (B.allocsz xor -1) );
      //FillChar( B3^, Sizeof( B3^ ), 0 );
      B3.smaller := nil;
      B3.larger := nil;
      B3.freesz := 0;
      B3.prev := B;
      B3.allocsz := -1;
    end;
    if B.allocsz = -1 then
    begin //     B
      {$IFDEF STATS}
      Inc( Allocated_FromEmpty );
      {$ENDIF}
      B.allocsz := size xor -1;
      Dec( B.freesz, size );
    end
      else
    begin //    B  ,    
      {$IFDEF STATS}
      Inc( Allocated_Dividing );
      {$ENDIF}
      cutsize := B.freesz div 2;
      if cutsize < size + Sizeof( B^ ) then
        cutsize := size + Sizeof( B^ );
      cutsize := (cutsize + 3) and not 3;
      B2 := Pointer( Integer( B ) + Sizeof( B^ ) + B.freesz + (B.allocsz xor -1) - cutsize );
      //FillChar( B2^, Sizeof( B2^ ), 0 );
      B2.smaller := nil;
      B2.larger := nil;
      //B2.freesz := nil;
      Dec( B.freesz, cutsize );
      //B2.next := B.next;
      //B.next := B2;
      B2.prev := B;
      {if B2.next <> nil then
        B2.next.prev := B2;}

      B2.freesz := cutsize - Sizeof( B^ ) - size;
      B2.allocsz := size xor -1;
      B3 := Pointer( Integer( B2 ) + Sizeof( B2^ ) + B2.freesz + (B2.allocsz xor -1) );
      B3.prev := B2;
      InsertSorted( B );
      B := B2;
    end;
    InsertSorted( B );
    Assert( log( 'alloc', B ) );
    Result := Pointer( DWORD( B ) + Sizeof( B^ ) );
  {$IFDEF USE_CRITICALSECTION}
  FINALLY
    LeaveCriticalSection( CritSec_FastMM );
  END;
  {$ENDIF USE_CRITICALSECTION}
end;
{$ENDIF NOASM_CODE}

//------------------------------------------------------------------------------
// Free memory block at given address
//------------------------------------------------------------------------------
{$IFDEF ASM_CODE}
function FreeMem_FastMM(p: Pointer): Integer;
const Bsz = Sizeof( TBlockInfo );
asm
  test eax, eax
  jz   @@fin
  push ebx
  lea  ebx, [eax-Bsz]
  {$IFDEF USE_CRITICALSECTION}
  push offset [CritSec_FastMM]
  call EnterCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  cmp  [ebx].TBlockInfo.allocsz, 0
  jl   @@1
  call dword ptr [OldMM.FreeMem]
  jmp  @@exit1
@@1:
  {$IFDEF STATS}
  inc  [Freed]
  {$ENDIF STATS}
  mov  eax, ebx
  call Unlink
  mov  eax, [ebx].TBlockInfo.allocsz
  not  eax
  add  [ebx].TBlockInfo.freesz, eax
  or   [ebx].TBlockInfo.allocsz, -1
  mov  ecx, [ebx].TBlockInfo.prev
  jecxz @@2
  push ecx
  xchg eax, ecx
  call Unlink
  pop  eax
  xchg ebx, eax
  mov  edx, [eax].TBlockInfo.freesz
  add  edx, Bsz
  add  [ebx].TBlockInfo.freesz, edx
  lea  edx, [ebx+Bsz]
  add  edx, [ebx].TBlockInfo.freesz
  mov  ecx, [ebx].TBlockInfo.allocsz
  not  ecx
  add  edx, ecx
  mov  [edx].TBlockInfo.prev, ebx

  mov  ecx, [ebx].TBlockInfo.prev
  or   ecx, [edx].TBlockInfo.freesz
  jnz  @@2
  mov  ecx, [ebx].TBlockInfo.allocsz
  and  ecx, [edx].TBlockInfo.allocsz
  inc  ecx
  jnz  @@2
  xchg eax, ebx
  call dword ptr [OldMM.FreeMem]
  jmp  @@exit1
@@2:
  xchg eax, ebx
  call InsertSorted
@@exit1:
  {$IFDEF USE_CRITICALSECTION}
  push offset [CritSec_FastMM]
  call LeaveCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  pop  ebx
@@fin:
  xor  eax, eax
end;
{$ELSE}
function FreeMem_FastMM(p: Pointer): Integer;
var B, B2, B3: PBlockInfo;
begin
  if p = nil then
  begin
    Result := 0;
    Exit;
  end;
  {$IFDEF USE_CRITICALSECTION}
  EnterCriticalSection( CritSec_FastMM );
  TRY
  {$ENDIF USE_CRITICALSECTION}
    B := Pointer( DWORD( p ) - Sizeof( B^ ) );
    if B.allocsz >= 0 then
    begin
      Result := OldMM.FreeMem( p );
      Exit;
    end;
    {$IFDEF STATS}
    Inc( Freed );
    {$ENDIF}
    Assert( log( 'free', B ) );
    Unlink( B );
    Inc( B.freesz, B.allocsz xor -1 );
    B.allocsz := -1;
    if B.prev <> nil then
    begin //     
      B2 := B;
      B := B.prev;
      Unlink( B );
      Inc( B.freesz, Sizeof( B^ ) + B2.freesz );
      //B.next := B2.next;
      //if B.next <> nil then
      //  B.next.prev := B;
      B3 := Pointer( Integer( B ) + Sizeof( B^ ) + B.freesz + (B.allocsz xor -1) );
      B3.prev := B;

      if (B.allocsz = -1) and (B.prev = nil) {and (B.next = nil)}
         and (B3.allocsz = -1) and (B3.freesz = 0) then
      begin
        OldMM.FreeMem( B );
        Result := 0;
        Exit;
      end;
    end;
    InsertSorted( B );
    Result := 0;
  {$IFDEF USE_CRITICALSECTION}
  FINALLY
    LeaveCriticalSection( CritSec_FastMM );
  END;
  {$ENDIF USE_CRITICALSECTION}
end;
{$ENDIF NOASM_CODE}

//------------------------------------------------------------------------------
// Reallocate memory block at given address to another given size
//------------------------------------------------------------------------------
{$IFDEF ASM_CODE}
function ReallocMem_FastMM(p: Pointer; size: Integer): Pointer;
const Bsz = Sizeof( TBlockInfo );
asm
  test eax, eax
  jnz  @@1
  xchg eax, edx
  jmp  GetMem_FastMM
@@1:
  push ebx
  push esi
  lea  ebx, [eax-Bsz]
  mov  esi, edx
  cmp  [ebx].TBlockInfo.allocsz, 0
  jl   @@2
  call dword ptr [OldMM.ReallocMem]
  jmp  @@fin
@@2:
  {$IFDEF USE_CRITICALSECTION}
  push offset[CritSec_FastMM]
  call EnterCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  mov  eax, [ebx].TBlockInfo.allocsz
  not  eax
  add  eax, [ebx].TBlockInfo.freesz
  sub  eax, esi
  jl   @@3
  {$IFDEF STATS}
  inc  [Reallocated]
  {$ENDIF STATS}
  push eax // new_freesz
  mov  eax, ebx
  call Unlink
  pop  [ebx].TBlockInfo.freesz
  mov  eax, esi
  not  eax
  mov  [ebx].TBlockInfo.allocsz, eax
  mov  eax, ebx
  call InsertSorted
  jmp  @@exit
@@3:
  {$IFDEF USE_CRITICALSECTION}
  push offset[CritSec_FastMM]
  call EnterCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  xchg eax, esi
  call GetMem_FastMM
  push eax
  push esi
  push edi
  xchg edi, eax
  lea  esi, [ebx+Bsz]
  mov  ecx, [ebx].TBlockInfo.allocsz
  not  ecx
  cmp  ecx, esi
  jle  @@4
  mov  ecx, esi
@@4:
  mov  eax, ecx
  shr  ecx, 2
  rep  movsd
  xchg ecx, eax
  and  ecx, 3
  rep  movsb
  pop  edi
  pop  esi
  pop  eax
  jmp  @@fin
@@exit:
  {$IFDEF USE_CRITICALSECTION}
  push offset[CritSec_FastMM]
  call EnterCriticalSection
  {$ENDIF USE_CRITICALSECTION}
  lea eax, [ebx+Bsz]
@@fin:
  pop esi
  pop ebx
end;
{$ELSE}
function ReallocMem_FastMM(p: Pointer; size: Integer): Pointer;
var B: PBlockInfo;
    new_freesz: Integer;
begin
  if p = nil then
  begin
    Result := GetMem_FastMM( size );
    Exit;
  end;
  B := nil;
  {$IFDEF USE_CRITICALSECTION}
  EnterCriticalSection( CritSec_FastMM );
  TRY
  {$ENDIF USE_CRITICALSECTION}
    B := Pointer( DWORD( p ) - Sizeof( B^ ) );
    if B.allocsz >= 0 then
    begin
      Result := OldMM.ReallocMem( p, size );
      Exit;
    end;
    if (B.allocsz xor -1) + B.freesz >= size then
    begin //     
      {$IFDEF STATS}
      Inc( Reallocated );
      {$ENDIF}
      Unlink( B );
      new_freesz := B.freesz + (B.allocsz xor -1) - size;
      B.freesz := new_freesz;
      B.allocsz := size xor -1;
      InsertSorted( B );
      Result := Pointer( DWORD( B ) + Sizeof( B^ ) );
      Assert( log( 'realloc', B ) );
      Exit;
    end;
  {$IFDEF USE_CRITICALSECTION}
  FINALLY
    LeaveCriticalSection( CritSec_FastMM );
  END;
  {$ENDIF USE_CRITICALSECTION}
  //   
  Result := GetMem_FastMM( size );
  Move( p^, Result^, Min( B.allocsz xor -1, size ) );
  FreeMem_FastMM( p );
end;
{$ENDIF NOASM_CODE}

initialization
  if FileExists( GetStartDir + 'FastMM.log' ) then
  Assert( DeleteFile( PChar( GetStartDir + 'FastMM.log' ) ) );

finalization

  if FastMM_Installed then
  begin
    Uninitialization := TRUE;
  end;

end.
